use std::{fs::File, path::Path};

use eyre::Context;
use log::{error, trace};

use crate::{Result, index::BagIndex, info::BagInfo, message::parse_msgs, reader::MmapReader};

pub struct Bag {
    reader: MmapReader,
    index: BagIndex,
}

impl Bag {
    pub fn open<P: AsRef<Path>>(path: P) -> Result<Self> {
        let bag_file = File::open(path).wrap_err("could not open bag file")?;
        let mut reader = MmapReader::map(bag_file).wrap_err("could not map bag file memory")?;
        let index = BagIndex::read(&mut reader).wrap_err("error reading bag index")?;
        Ok(Bag { reader, index })
    }

    pub fn compute_message_layouts(&mut self) -> Result<()> {
        for conn in &self.index.connections {
            match parse_msgs(conn) {
                Ok(msgs) => {
                    for msg in &msgs {
                        trace!(
                            "message definition parsed: {:#?}",
                            msg.fields()
                                .iter()
                                .filter(|field| !field.is_constant())
                                .map(ToString::to_string)
                                .collect::<Vec<_>>()
                        );
                    }
                }
                Err(err) => error!("could not parse message definition: {}", err),
            }
        }
		Ok(())
    }

	pub fn compute_info(&mut self) -> Result<BagInfo> {
		BagInfo::compute(&mut self.reader, &self.index)
	}
}

