use std::{io, str::FromStr};

use eyre::bail;

use crate::{
    error,
    parse::{Header, Op, Time},
    reader::BagReader,
    Error, Result,
};

#[derive(Clone, Copy, Debug)]
pub enum Compression {
    None,
    Bz2,
    Lz4,
}

impl FromStr for Compression {
    type Err = Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(match s {
            "none" => Compression::None,
            "bz2" => Compression::Bz2,
            "lz4" => Compression::Lz4,
            _ => bail!(error::UnsupportedCompression(s.to_string())),
        })
    }
}

impl Compression {
    pub fn decompress_stream<'a, R: io::Read + 'a>(self, read: R) -> Box<dyn io::Read + 'a> {
        match self {
            Compression::None => Box::new(read),
            Compression::Bz2 => todo!("bz2 decompression"),
            Compression::Lz4 => Box::new(lz4_flex::frame::FrameDecoder::new(read)),
        }
    }
}

#[derive(Debug)]
pub struct ChunkHeader {
    pub compression: Compression,
    pub uncompressed_size: u32,
}

impl ChunkHeader {
    pub fn read<R: BagReader>(reader: &mut R) -> Result<Self> {
        let header = reader.read_header_op(Op::Chunk)?;
        Self::from_header(header)
    }

    pub fn from_header(header: Header) -> Result<Self> {
        Ok(Self {
            compression: header.read_string(b"compression")?.parse()?,
            uncompressed_size: header.read_u32(b"size")?,
        })
    }
}

pub fn read_chunk<R: BagReader>(bag_reader: &mut R, pos: u64) -> Result<Vec<u8>> {
    bag_reader.seek(io::SeekFrom::Start(pos))?;

    let chunk_header = ChunkHeader::read(bag_reader)?;

    let compressed_data = bag_reader.read_data()?;
    let mut data = Vec::with_capacity(chunk_header.uncompressed_size as usize);
    let mut decompresor = chunk_header.compression.decompress_stream(compressed_data);
    io::copy(&mut decompresor, &mut data)?;

    Ok(data)
}

pub struct MessageDataHeader {
    pub conn_id: u32,
    pub time: Time,
}

impl MessageDataHeader {
    pub fn read<R: BagReader>(reader: &mut R) -> Result<Self> {
        let header = reader.read_header_op(Op::MsgData)?;
        Self::from_header(header)
    }

    pub fn from_header(header: Header) -> Result<Self> {
        Ok(Self {
            conn_id: header.read_u32(b"conn")?,
            time: header.read_time(b"time")?,
        })
    }
}
