use std::collections::HashMap;

use eyre::{bail, Context};
use rayon::prelude::*;

use crate::{
    chunk::{read_chunk, MessageDataHeader},
    index::BagIndex,
    parse::Op,
    reader::{BagReader, SliceReader},
    Result,
};

#[derive(Default, Debug)]
pub struct BagInfo {
    pub total_uncompressed: u64,
    pub per_connection: HashMap<u32, u64>,
}

impl BagInfo {
    fn combine(mut self, other: BagInfo) -> BagInfo {
        self.total_uncompressed += other.total_uncompressed;
        for (conn, count) in other.per_connection {
            *self.per_connection.entry(conn).or_insert(0) += count;
        }
        self
    }

    pub fn compute<R: BagReader + Clone + Send + Sync>(reader: &mut R, index: &BagIndex) -> Result<BagInfo> {
        index
            .chunks
            .par_iter()
            // .try_fold(BagInfo::default, |mut info, chunk| -> rsbag::Result<_> {
            //     let mut reader = reader.clone();
            //     reader.seek(io::SeekFrom::Start(chunk.pos))?;
            //     let chunk_header = ChunkHeader::read(&mut reader)?;
            //     info.total_uncompressed += chunk_header.uncompressed_size as u64;
            //     reader.skip_data()?;
            //     for _ in &chunk.connections {
            //         let index = IndexData::read(&mut reader)?;
            //         *info.per_connection.entry(index.conn_id).or_insert(0) +=
            //             index.entries.len() as u64;
            //     }
            //     Ok(info)
            // })
            .try_fold(BagInfo::default, |mut info, chunk| -> Result<_> {
                let mut reader = reader.clone();
                let data = read_chunk(&mut reader, chunk.pos)
                    .wrap_err_with(|| format!("failed to read chunk: {:#?}", chunk))?;
                info.total_uncompressed += data.len() as u64;
                let mut chunk_reader = SliceReader::from(data);
                while chunk_reader.remaining() > 0 {
                    let header = chunk_reader.read_header()?;
                    let op = header.read_op()?;
                    match op {
                        Op::MsgData => {
                            let header = MessageDataHeader::from_header(header)?;
                            let count = info.per_connection.entry(header.conn_id).or_insert(0);
                            *count += 1;
                            chunk_reader.skip_data()?;
                        }
                        Op::Connection => chunk_reader.skip_data()?,
                        _ => bail!("unexpected op in chunk: {:?}", op),
                    }
                }
                Ok(info)
            })
            .try_reduce(BagInfo::default, |a, b| Ok(a.combine(b)))
    }
}
