use nom::{
    combinator::all_consuming,
    number::complete::{le_u32, le_u64, le_u8},
    Parser,
};
use num_enum::TryFromPrimitive;

use crate::parse::IResult;

use super::{FieldDataError, Input, Result};

#[derive(Clone, Copy, Debug, TryFromPrimitive, PartialEq)]
#[repr(u8)]
pub enum Op {
    MsgData = 0x02,
    FileHeader = 0x03,
    IndexData = 0x04,
    Chunk = 0x05,
    ChunkInfo = 0x06,
    Connection = 0x07,
}

impl Op {
    pub fn read(input: Input) -> Result<Self> {
        let (_, op) = all_consuming(le_u8)
            .parse(input)
            .map_err(FieldDataError::from)?;
        Op::try_from_primitive(op).map_err(Into::into)
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord)]
pub struct Time {
    pub sec: u32,
    pub nsec: u32,
}

impl Time {
    pub fn from_packed(packed: u64) -> Time {
        Time {
            sec: packed as u32,
            nsec: (packed >> 32) as u32,
        }
    }

    pub fn parse(input: Input) -> IResult<Self> {
        le_u64.map(Time::from_packed).parse(input)
    }

    pub fn read(input: Input) -> Result<Self> {
        let (_, x) = Self::parse(input).map_err(FieldDataError::from)?;
        Ok(x)
    }
}

pub fn read_u32(input: Input) -> Result<u32> {
    let (_, x) = all_consuming(le_u32)
        .parse(input)
        .map_err(FieldDataError::from)?;
    Ok(x)
}

pub fn read_u64(input: Input) -> Result<u64> {
    let (_, x) = all_consuming(le_u64)
        .parse(input)
        .map_err(FieldDataError::from)?;
    Ok(x)
}

pub fn read_string(input: Input) -> Result<String> {
    Ok(String::from_utf8(input.to_owned())?)
}
