use std::io::SeekFrom;

use nom::multi::length_data;
use nom::number::streaming::le_u32;

use crate::error::Result;
use crate::parse::{self, Header, Op, Version};

pub mod error;
mod io;
#[cfg(feature = "mmap")]
mod mmap;
mod slice;

#[cfg(feature = "mmap")]
pub use self::mmap::MmapReader;
pub use self::{io::IoReader, slice::SliceReader};

pub trait BagReader {
    fn read_parser<'a, O: 'a, P>(&'a mut self, parser: P) -> Result<O>
    where
        P: nom::Parser<&'a [u8], O, parse::Error<&'a [u8]>>;

    fn seek(&mut self, pos: SeekFrom) -> Result<()>;

    fn read_version(&mut self) -> Result<Version> {
        self.read_parser(Version::parse)
    }

    fn read_header(&mut self) -> Result<Header> {
        self.read_parser(Header::parse)
    }

    fn read_header_op(&mut self, op: Op) -> Result<Header> {
        let header = self.read_header()?;
        let actual_op = header.read_op()?;
        if actual_op != op {
            return Err(error::UnexpectedOp(op, actual_op).into());
        }
        Ok(header)
    }

    fn read_data_length(&mut self) -> Result<u32> {
        self.read_parser(le_u32)
    }

    fn skip_data(&mut self) -> Result<()> {
        let data_length = self.read_data_length()?;
        self.seek(SeekFrom::Current(data_length as i64))?;
        Ok(())
    }

    fn read_data(&mut self) -> Result<&[u8]> {
        self.read_parser(length_data(le_u32))
    }
}
